import contextlib
import ldap
import logging
from ldap.dn import escape_dn_chars
from ldap.filter import escape_filter_chars
from ldap.ldapobject import LDAPObject
from authserv import model


class Error(Exception):
    pass


class UserDb(model.UserDb):

    ldap_attrs = [
        'userPassword',
        'totpSecret',
        'appSpecificPassword',
        ]

    def __init__(self, service_map, ldap_uri, ldap_bind_dn, ldap_bind_pw):
        self.service_map = service_map
        self.ldap_uri = ldap_uri
        self.ldap_bind_dn = ldap_bind_dn
        self.ldap_bind_pw = ldap_bind_pw

    @contextlib.contextmanager
    def _conn(self):
        c = LDAPObject(self.ldap_uri)
        c.protocol_version = ldap.VERSION3
        c.simple_bind_s(self.ldap_bind_dn, self.ldap_bind_pw)
        yield c
        c.unbind_s()

    def _query_user(self, username, service):
        # Allow referencing a service to another, by specifying a
        # string rather than a dictionary as the value. If you build
        # infinite loops this way, it's your fault.
        ldap_params = self.service_map.get(service)
        while isinstance(ldap_params, basestring):
            ldap_params = self.service_map.get(ldap_params)
        if not ldap_params:
            logging.error('unknown service "%s"', service)
            return None

        with self._conn() as c:
            # LDAP queries can be built in two ways:
            #
            # - specify a fully-qualified 'dn' attribute, in which
            # case we'll run a SCOPE_BASE query for that specific DN.
            # In this case, 'filter' is optional.
            #
            # - specify 'base' and 'filter' together to identify a
            # single object. This is a SCOPE_SUBTREE search and
            # 'filter' is required.
            #
            if 'dn' in ldap_params:
                base = ldap_params['dn'].replace('%s', escape_dn_chars(username))
                filt = ldap_params.get('filter', '(objectClass=*)').replace('%s', escape_filter_chars(username))
                scope = ldap.SCOPE_BASE
            else:
                base = ldap_params['base'].replace('%s', escape_dn_chars(username))
                filt = ldap_params['filter'].replace('%s', escape_filter_chars(username))
                scope = ldap.SCOPE_SUBTREE
            logging.debug('ldap search: base=%s, scope=%s, filt=%s', base, scope, filt)
            result = c.search_s(base, scope, filt, self.ldap_attrs)

            if not result:
                return None
            if len(result) > 1:
                raise Error('too many results from LDAP')

            return User(username, result[0][0], result[0][1])

    def get_user(self, username, service):
        try:
            return self._query_user(username, service)
        except (Error, ldap.LDAPError), e:
            logging.error('userdb error: %s', e)
            return None


class User(model.User):

    def __init__(self, username, dn, data):
        self._username = username
        self._dn = dn
        self._otp_enabled = False
        self._asps = []
        for key, values in data.iteritems():
            if key == 'userPassword':
                self._password = values[0]
                if self._password.startswith('{crypt}'):
                    self._password = self._password[7:]
            elif key == 'totpSecret':
                self._otp_enabled = True
                self._totp_secret = values[0]
            elif key == 'appSpecificPassword':
                self._asps = [v.split(':', 1) for v in values]

    def otp_enabled(self):
        return self._otp_enabled

    def app_specific_passwords_enabled(self):
        return bool(self._asps)

    def get_name(self):
        return self._username

    def get_totp_key(self):
        return self._totp_secret

    def get_app_specific_passwords(self, service):
        return [x[1] for x in self._asps if x[0] == service]

    def get_password(self):
        return self._password
