import functools
from flask import abort, request
from authserv import app
from authserv import protocol


key_sep = ':'


def key_from_request(header=None, param=None):
    """Build a key from HTTP request headers and query arguments."""
    if not header and not param:
        raise Exception('Must set either header or param')

    required_parts = 2
    if header:
        required_parts += 1
    if param:
        required_parts += 1

    def _key(prefix, args):
        parts = [request.method, request.path]
        if header:
            value = request.environ.get(header)
            if value:
                parts.append('%s=%s' % (header, value))
        if param:
            if request.method == 'POST':
                value = request.form.get(param)
            else:
                value = request.args.get(param)
            if value:
                parts.append('%s=%s' % (param, value))
        if len(parts) >= required_parts:
            return prefix + key_sep + key_sep.join(parts)
        return None
    return _key


def key_from_args(*args_idx):
    """Build a key from the wrapped function's arguments.

    This function should be passed the positional index of the
    arguments that will be used to build the key.
    """
    def _key(prefix, args):
        parts = [args[x] for x in args_idx if args[x]]
        if not parts:
            return None
        return prefix + key_sep + key_sep.join(parts)
    return _key


class RateLimit(object):
    """Keep track of request rates using Memcache."""

    prefix = 'authserv/r:'

    def __init__(self, count, period):
        self.count = count
        self.period = period

    def check(self, mc, key):
        key = self.prefix + key
        try:
            result = mc.incr(key)
        except:
            result = None
        if result is None:
            result = 1
            if not mc.add(key, result, time=self.period):
                try:
                    result = mc.incr(key)
                except:
                    result = None
                if result is None:
                    # Memcache is failing.
                    return True
        return result <= self.count


def ratelimit_http_request(key_fn, count=0, period=0):
    """Rate limit an HTTP request handler.

    If the rate limit is triggered, the request will generate a HTTP
    503 error.
    """

    rl = RateLimit(count, period)

    def decoratorfn(fn):
        @functools.wraps(fn)
        def _ratelimit(*args, **kwargs):
            key = key_fn(fn.__name__, args)
            if key and not rl.check(app.memcache, key):
                app.logger.debug('ratelimited: %s', key)
                abort(503)
            return fn(*args, **kwargs)
        return _ratelimit
    return decoratorfn


class BlackList(object):
    """Block requests once a rate limit is triggered."""

    prefix = 'authserv/b:'

    def __init__(self, count, period, ttl):
        self.rl = RateLimit(count, period)
        self.ttl = ttl

    def check(self, mc, key):
        key = self.prefix + key
        result = mc.get(key)
        return result is None

    def incr(self, mc, key):
        key = self.prefix + key
        if not self.rl.check(mc, key):
            mc.set(key, 'true', time=self.ttl)


def blacklist_on_auth_failure(key_fn, count=0, period=0, ttl=0):
    """Blacklist authentication failures.

    The wrapped function should return one of the error codes from
    protocol.py. If it returns ERR_AUTHENTICATION_FAILURE more often
    than the specified threshold, the request parameters are added to
    the blacklist.

    When a request is blacklisted, this function will return
    ERR_AUTHENTICATION_FAILURE.
    """

    bl = BlackList(count, period, ttl)

    def decoratorfn(fn):
        @functools.wraps(fn)
        def _blacklist(*args, **kwargs):
            key = key_fn(fn.__name__, args)
            if not key:
                return fn(*args, **kwargs)

            if not bl.check(app.memcache, key):
                app.logger.debug('blacklisted %s', key)
                return protocol.ERR_AUTHENTICATION_FAILURE
            result = fn(*args, **kwargs)
            if result != protocol.OK:
                bl.incr(app.memcache, key)
            return result
        return _blacklist
    return decoratorfn

